@@ -71,7 +71,7 @@ class MemoryFilesystemTest(test.TestCase):
         self.assertEqual('c', fs['/c'])
         self.assertEqual('db', fs['/d/b'])
 
-    def test_ls_recursive(self):
+    def test_old_ls_recursive(self):
         fs = impl_memory.FakeFilesystem()
         fs.ensure_path("/d")
         fs.ensure_path("/c/d")
@@ -91,7 +91,65 @@ class MemoryFilesystemTest(test.TestCase):
             '/a/b/c/d',
         ], contents)
 
+    def test_ls_recursive(self):
+        fs = impl_memory.FakeFilesystem()
+        fs.ensure_path("/d")
+        fs.ensure_path("/c/d")
+        fs.ensure_path("/b/c/d")
+        fs.ensure_path("/a/b/c/d")
+        contents = fs.ls_r("/", absolute=False)
+        self.assertEqual([
+            'a',
+            'b',
+            'c',
+            'd',
+            'a/b',
+            'b/c',
+            'c/d',
+            'a/b/c',
+            'b/c/d',
+            'a/b/c/d',
+        ], contents)
+
+    def test_ls_recursive_absolute(self):
+        fs = impl_memory.FakeFilesystem()
+        fs.ensure_path("/d")
+        fs.ensure_path("/c/d")
+        fs.ensure_path("/b/c/d")
+        fs.ensure_path("/a/b/c/d")
+        contents = fs.ls_r("/", absolute=True)
+        self.assertEqual([
+            '/a',
+            '/b',
+            '/c',
+            '/d',
+            '/a/b',
+            '/b/c',
+            '/c/d',
+            '/a/b/c',
+            '/b/c/d',
+            '/a/b/c/d',
+        ], contents)
+
     def test_ls_recursive_targeted(self):
+        fs = impl_memory.FakeFilesystem()
+        fs.ensure_path("/d")
+        fs.ensure_path("/c/d")
+        fs.ensure_path("/b/c/d")
+        fs.ensure_path("/a/b/c/d")
+        contents = fs.ls_r("/a/b", absolute=False)
+        self.assertEqual(['c', 'c/d'], contents)
+
+    def test_ls_recursive_targeted_absolute(self):
+        fs = impl_memory.FakeFilesystem()
+        fs.ensure_path("/d")
+        fs.ensure_path("/c/d")
+        fs.ensure_path("/b/c/d")
+        fs.ensure_path("/a/b/c/d")
+        contents = fs.ls_r("/a/b", absolute=True)
+        self.assertEqual(['/a/b/c', '/a/b/c/d'], contents)
+
+    def test_old_ls_recursive_targeted_absolute(self):
         fs = impl_memory.FakeFilesystem()
         fs.ensure_path("/d")
         fs.ensure_path("/c/d")
@@ -128,6 +186,12 @@ class MemoryFilesystemTest(test.TestCase):
         fs = impl_memory.FakeFilesystem()
         self.assertRaises(exc.NotFound, self._get_item_path, fs, '/c')
 
+    def test_bad_norms(self):
+        fs = impl_memory.FakeFilesystem()
+        self.assertRaises(ValueError, fs.normpath, '')
+        self.assertRaises(ValueError, fs.normpath, 'abc/c')
+        self.assertRaises(ValueError, fs.normpath, '../c')
+
     def test_del_root_not_allowed(self):
         fs = impl_memory.FakeFilesystem()
         self.assertRaises(ValueError, self._del_item_path, fs, '/')
